==> materia-core-0.3.0-20150613.153231-20.jar: materia/middleware_set.clj <==
(ns materia.middleware-set
  (:require [clojure.tools.namespace.dependency :as dep])
  (:import [clojure.tools.namespace.dependency
            DependencyGraphUpdate DependencyGraph]))

(defprotocol Wrappable
  (wrap [this h]))

(defprotocol Identifiable
  (get-key [this]))

(extend-protocol Wrappable
  clojure.lang.AFn
  (wrap [this h]
    (this h)))

(extend-protocol Identifiable
  clojure.lang.AFn
  (get-key [this]
    this)
  clojure.lang.Keyword
  (get-key [this]
    this))

(defrecord Middleware [tag f args]
  Wrappable
  (wrap [this h]
    (apply f h args))

  Identifiable
  (get-key [this]
    tag))

(defrecord DisabledMiddleware [tag f args]
  Wrappable
  (wrap [this h]
    h)

  Identifiable
  (get-key [this]
    tag))

(defprotocol MiddlewareSet
  (add [this middleware] [this middleware deps])
  (get-all [this]))

(defrecord Middlewares [m dep-graph]
  DependencyGraph
  (immediate-dependencies [this node]
    (dep/immediate-dependencies dep-graph node))
  (immediate-dependents [this node]
    (dep/immediate-dependents dep-graph node))
  (transitive-dependencies [this node]
    (dep/transitive-dependencies dep-graph node))
  (transitive-dependents [this node]
    (dep/transitive-dependents dep-graph node))
  (nodes [this]
    (dep/nodes dep-graph))

  DependencyGraphUpdate
  (depend [this node dep]
    (Middlewares. m (dep/depend dep-graph (get-key node) (get-key dep))))
  (remove-edge [this node dep]
    (Middlewares. m (dep/remove-edge dep-graph (get-key node) (get-key dep))))
  (remove-all [this node]
    (Middlewares. m (dep/remove-all dep-graph (get-key node))))
  (remove-node [this node]
    (Middlewares. m (dep/remove-node dep-graph (get-key node))))

  Wrappable
  (wrap [this h]
    (reduce (fn [acc middleware]
              (if (satisfies? Wrappable middleware)
                (wrap middleware acc)
                acc))
            h
            (reverse (get-all this))))

  MiddlewareSet
  (add [this middleware]
    (add this middleware []))
  (add [this middleware deps]
    (let [key  (get-key middleware)
          deps (conj deps ::root)] ; Should apply also independent middlewares
      (Middlewares. (assoc m key middleware)
                    (reduce #(dep/depend %1 key %2) dep-graph deps))))
  (get-all [this]
    (filter some? (map #(get m %) (dep/topo-sort dep-graph)))))

(defn middlewares []
  (->Middlewares {} (dep/graph)))

(defn middleware [tag f & args]
  (->Middleware tag f args))

(defn disabled-middleware [tag f & args]
  (->DisabledMiddleware tag f args))

(defn add-middleware
  ([bag f]
   (add bag f))
  ([bag f deps]
   (add bag f deps))
  ([bag key f deps]
   (add bag (middleware key f) deps))
  ([bag key f deps & args]
   (add bag (middleware key f args) deps)))

(defn append [bag m]
  (add bag m (dep/nodes bag)))

(defn preped [bag m]
  (reduce (fn [acc k]
            (dep/depend acc (get-key k) m))
          (add bag m)
          (get-all bag)))

(defn after [bag anchor m]
  (-> bag
      (add m)
      (dep/depend m anchor)))

(defn before [bag anchor m]
  (-> bag
      (add m)
      (dep/depend anchor m)))

(defn between [bag from to m]
  (-> bag
      (add m)
      (dep/depend m from)
      (dep/depend to m)))

(defn add-linearly [bag ms & [init-deps]]
  (->> ms
       (mapv (comp vector get-key))
       (cons init-deps)
       (map vector ms)
       (reduce (fn [acc [m deps]]
                 (add acc m deps))
               bag)))

==> materia-core-0.3.0-20150627.125429-37.jar: materia/middleware_set.clj <==
(ns materia.middleware-set
  (:require [clojure.tools.namespace.dependency :as dep])
  (:import [clojure.tools.namespace.dependency
            DependencyGraphUpdate DependencyGraph]))

(defprotocol Wrappable
  (wrap [this h]))

(defprotocol Identifiable
  (get-key [this]))

(extend-protocol Wrappable
  clojure.lang.AFn
  (wrap [this h]
    (this h)))

(extend-protocol Identifiable
  clojure.lang.AFn
  (get-key [this]
    this)
  clojure.lang.Keyword
  (get-key [this]
    this))

(defrecord Middleware [tag f args]
  Wrappable
  (wrap [this h]
    (apply f h args))

  Identifiable
  (get-key [this]
    tag))

(defrecord DisabledMiddleware [tag f args]
  Wrappable
  (wrap [this h]
    h)

  Identifiable
  (get-key [this]
    tag))

(defprotocol MiddlewareSet
  (add [this middleware] [this middleware deps])
  (get-all [this]))

(defrecord Middlewares [m dep-graph]
  DependencyGraph
  (immediate-dependencies [this node]
    (dep/immediate-dependencies dep-graph node))
  (immediate-dependents [this node]
    (dep/immediate-dependents dep-graph node))
  (transitive-dependencies [this node]
    (dep/transitive-dependencies dep-graph node))
  (transitive-dependents [this node]
    (dep/transitive-dependents dep-graph node))
  (nodes [this]
    (dep/nodes dep-graph))

  DependencyGraphUpdate
  (depend [this node dep]
    (Middlewares. m (dep/depend dep-graph (get-key node) (get-key dep))))
  (remove-edge [this node dep]
    (Middlewares. m (dep/remove-edge dep-graph (get-key node) (get-key dep))))
  (remove-all [this node]
    (Middlewares. m (dep/remove-all dep-graph (get-key node))))
  (remove-node [this node]
    (Middlewares. m (dep/remove-node dep-graph (get-key node))))

  Wrappable
  (wrap [this h]

    (reduce (fn [acc middleware]
              (if (satisfies? Wrappable middleware)
                (wrap middleware acc)
                acc))
            h
            (reverse (get-all this))))

  MiddlewareSet
  (add [this middleware]
    (add this middleware []))
  (add [this middleware deps]
    (let [key  (get-key middleware)
          deps (conj deps ::root)] ; Should apply also independent middlewares
      (Middlewares. (assoc m key middleware)
                    (reduce #(dep/depend %1 key %2) dep-graph deps))))
  (get-all [this]
    (filter some? (map #(get m %) (dep/topo-sort dep-graph)))))

(defn middlewares []
  (->Middlewares {} (dep/graph)))

(defn middleware [tag f & args]
  (->Middleware tag f args))

(defn disabled-middleware [tag f & args]
  (->DisabledMiddleware tag f args))

(defn add-middleware
  ([bag f]
   (add bag f))
  ([bag f deps]
   (add bag f deps))
  ([bag key f deps]
   (add bag (middleware key f) deps))
  ([bag key f deps & args]
   (add bag (apply middleware key f args) deps)))

(defn append [bag m]
  (add bag m (dep/nodes bag)))

(defn prepend [bag m]
  (reduce (fn [acc k]
            (dep/depend acc (get-key k) m))
          (add bag m)
          (get-all bag)))

(defn after [bag anchor m]
  (-> bag
      (add m)
      (dep/depend m anchor)))

(defn before [bag anchor m]
  (-> bag
      (add m)
      (dep/depend anchor m)))

(defn between [bag from to m]
  (-> bag
      (add m)
      (dep/depend m from)
      (dep/depend to m)))

(defn add-linearly [bag ms & [init-deps]]
  (->> ms
       (mapv (comp vector get-key))
       (cons init-deps)
       (map vector ms)
       (reduce (fn [acc [m deps]]
                 (add acc m deps))
               bag)))
